/******************************************************************************
 *
 * Copyright (C) 1996-2001 Jason Evans <jasone@canonware.com>.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice(s), this list of conditions and the following disclaimer
 *    unmodified other than the allowable addition of one or more
 *    copyright notices.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice(s), this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER(S) ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER(S) BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 ******************************************************************************
 *
 * Version: 2.0.0
 *
 ******************************************************************************/

typedef struct cw_nxo_s cw_nxo_t;
typedef struct cw_nxoe_s cw_nxoe_t;

/* Declared here to avoid circular header dependencies. */
typedef struct cw_nx_s cw_nx_t;
typedef struct cw_nxa_s cw_nxa_t;
typedef void cw_op_t(cw_nxo_t *);

/* Interpreter errors. */
typedef enum {
	NXO_THREADE_NONE,		/* No error. */
	NXO_THREADE_DSTACKUNDERFLOW,	/* No poppable dictionary on dstack. */
	NXO_THREADE_ESTACKOVERFLOW,	/* estack too deep. */
	NXO_THREADE_INVALIDACCESS,	/* Permission error. */
	NXO_THREADE_INVALIDEXIT,	/* exit operator called outside loop. */
	NXO_THREADE_INVALIDFILEACCESS,	/* Insufficient file permissions. */
	NXO_THREADE_IOERROR,		/* read()/write()/etc. error. */
	NXO_THREADE_LIMITCHECK,		/* Value outside legal range. */
	NXO_THREADE_RANGECHECK,		/* Out of bounds string or array use. */
	NXO_THREADE_STACKUNDERFLOW,	/* Not enough objects on ostack. */
	NXO_THREADE_SYNTAXERROR,	/* Scanner syntax error. */
	NXO_THREADE_TYPECHECK,		/* Incorrect argument type. */
	NXO_THREADE_UNDEFINED,		/* Object not found in dstack. */
	NXO_THREADE_UNDEFINEDFILENAME,	/* Bad filename. */
	NXO_THREADE_UNDEFINEDRESULT,	/* Divide by 0. */
	NXO_THREADE_UNMATCHEDFINO,	/* No fino on ostack. */
	NXO_THREADE_UNMATCHEDMARK,	/* No mark on ostack. */
	NXO_THREADE_UNREGISTERED	/* Other non-enumerated error. */
#define	NXO_THREADE_LAST	NXO_THREADE_UNREGISTERED
} cw_nxo_threade_t;

typedef enum {
	NXOT_NO,
	NXOT_ARRAY,
	NXOT_BOOLEAN,
#ifdef _CW_THREADS
	NXOT_CONDITION,
#endif
	NXOT_DICT,
	NXOT_FILE,
	NXOT_FINO,
	NXOT_HOOK,
	NXOT_INTEGER,
	NXOT_MARK,
#ifdef _CW_THREADS
	NXOT_MUTEX,
#endif
	NXOT_NAME,
	NXOT_NULL,
	NXOT_OPERATOR,
	NXOT_PMARK,
	NXOT_STACK,
	NXOT_STRING,
	NXOT_THREAD
#define	NXOT_LAST	NXOT_THREAD
}	cw_nxot_t;

/* Attribute. */
typedef enum {
	NXOA_LITERAL,
	NXOA_EXECUTABLE,
	NXOA_EVALUATABLE
}	cw_nxoa_t;

typedef cw_sint64_t cw_nxoi_t;

/*
 * Main object structure.
 */
struct cw_nxo_s {
#ifdef _CW_DBG
	cw_uint32_t	magic;
#define _CW_NXO_MAGIC	0x398754ba
#endif

	/*
	 * We can't use bit fields here, since we need explicit knowledge of
	 * structure layout to avoid GC races in nxo_dup() and nxo_p_new().
	 * Therefore, use a single 32 bit variable and do the bit manipulation
	 * manually.
	 *
	 * . : Unused.
	 *
	 * C : Op code.  If this is an operator, and NXN_ZERO < op code <=
	 *     NXN_LAST, op code corresponds to the name of this operator.  This
	 *     can be used to print the operator name or inline fast_op's in the
	 *     interpreter loop.
	 *
	 * B : Array bound.  TRUE if the bind operator has processed this array.
	 *     This is used to avoid infinite recursion in the bind operator
	 *     when binding recursive procedures.
	 *
	 * A : Attribute.  An nxo is LITERAL, EXECUTABLE, or EVALUATABLE.
	 *
	 * P : Fast op.  If type is NXOT_OPERATOR and fast op is TRUE, this
	 *     operator can be handled specially in nxo_thread_loop().
	 *
	 * T : Type.
	 *
	 * ........ .....CCC CCCCCCCB AAPTTTTT
	 *
	 */
	cw_uint32_t	flags;

	union {
		struct {
			cw_bool_t	val;
		}	boolean;
		struct {
			cw_nxoi_t	i;
		}	integer;
		struct {
			cw_op_t		*f;
		}	operator;
		cw_nxoe_t	*nxoe;
	}	o;
};

/*
 * All extended type objects contain a nxoe.  This provides a poor man's
 * inheritance.  Since onyx's type system is non-extenible, this idiom is
 * adequate.
 */
struct cw_nxoe_s {
#ifdef _CW_DBG
	cw_uint32_t	magic;
#define _CW_NXOE_MAGIC	0x0fa6e798
#endif

	/*
	 * Linkage for GC.  All nxoe's are in a single ring, which the GC uses
	 * to implement a Baker's Treadmill collector.
	 */
	qr(cw_nxoe_t)	link;
	/*
	 * Object type.  We store this in nxoe's as well as nxo's, since various
	 * functions access nxoe's directly, rather than going through a
	 * referring nxo.
	 */
	cw_nxot_t	type:5;
	/*
	 * If TRUE, the string in the key is statically allocated, and should
	 * not be deallocated during destruction.
	 */
	cw_bool_t	name_static:1;
	/*
	 * The GC toggles this value at each collection in order to maintain
	 * state.
	 */
	cw_bool_t	color:1;
	/*
	 * TRUE if this object has been registered with the GC.  It is possible
	 * for an object to be reachable by the GC (on a stack, for instance),
	 * but not be registered yet.
	 */
	cw_bool_t	registered:1;
#ifdef _CW_THREADS
	/*
	 * If true, accesses to this object are locked.  This applies to arrays,
	 * dictionaries, files, and strings.
	 */
	cw_bool_t	locking:1;
#endif
	/*
	 * If TRUE, this nxoe is a reference to another nxoe.
	 */
	cw_bool_t	indirect:1;
};

cw_sint32_t nxo_compare(cw_nxo_t *a_a, cw_nxo_t *a_b);
cw_nxoe_t *nxo_nxoe_get(cw_nxo_t *a_nxo);
#ifdef _CW_THREADS
cw_bool_t nxo_lcheck(cw_nxo_t *a_nxo);
#endif

#ifndef _CW_USE_INLINES
void	nxo_dup(cw_nxo_t *a_to, cw_nxo_t *a_from);
cw_nxot_t nxo_type_get(cw_nxo_t *a_nxo);
cw_nxoa_t nxo_attr_get(cw_nxo_t *a_nxo);
void	nxo_attr_set(cw_nxo_t *a_nxo, cw_nxoa_t a_attr);
void	nxo_p_new(cw_nxo_t *a_nxo, cw_nxot_t a_type);
#endif

#if (defined(_CW_USE_INLINES) || defined(_NXO_C_))
_CW_INLINE void
nxo_dup(cw_nxo_t *a_to, cw_nxo_t *a_from)
{
	_cw_check_ptr(a_to);
	_cw_dassert(a_to->magic == _CW_NXO_MAGIC);

	_cw_check_ptr(a_from);
	_cw_dassert(a_from->magic == _CW_NXO_MAGIC);

	/*
	 * The order of operations is important in order to avoid a GC race.
	 */
	a_to->flags = 0;
	a_to->o = a_from->o;
	a_to->flags = a_from->flags;
}

_CW_INLINE cw_nxot_t
nxo_type_get(cw_nxo_t *a_nxo)
{
	_cw_check_ptr(a_nxo);
	_cw_dassert(a_nxo->magic == _CW_NXO_MAGIC);

	return ((cw_nxot_t)(a_nxo->flags & 0x1f));
}

_CW_INLINE cw_nxoa_t
nxo_attr_get(cw_nxo_t *a_nxo)
{
	_cw_check_ptr(a_nxo);
	_cw_dassert(a_nxo->magic == _CW_NXO_MAGIC);

	return ((cw_nxoa_t)(((a_nxo)->flags >> 6) & 3));
}

_CW_INLINE void
nxo_attr_set(cw_nxo_t *a_nxo, cw_nxoa_t a_attr)
{
	_cw_check_ptr(a_nxo);
	_cw_dassert(a_nxo->magic == _CW_NXO_MAGIC);

	a_nxo->flags = (a_nxo->flags & 0xffffff3f) | (a_attr << 6);
}

/*
 * Private, but various object constructor macros need its definition.
 */
_CW_INLINE void
nxo_p_new(cw_nxo_t *a_nxo, cw_nxot_t a_type)
{
	/*
	 * The order of operations is important in order to avoid a GC race.
	 */
	a_nxo->flags = 0;
#ifdef _CW_DBG
	a_nxo->magic = _CW_NXO_MAGIC;
#endif

	/*
	 * o.integer.i is assumed to be at least as big as all the other fields
	 * in the union.
	 */
	a_nxo->o.integer.i = 0;
	a_nxo->flags = a_type;
}
#endif	/* (defined(_CW_USE_INLINES) || defined(_NXO_C_)) */
